@@ -1,15 +1,17 @@
 import json
 from abc import ABC, abstractmethod
-from threading import RLock
 from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union
 
 import pandas as pd
 import pyarrow as pa
+from fugue.collections.yielded import Yielded
 from fugue.exceptions import FugueDataFrameEmptyError, FugueDataFrameOperationError
+from triad import SerializableRLock
 from triad.collections.dict import ParamDict
 from triad.collections.schema import Schema
 from triad.exceptions import InvalidOperationError
 from triad.utils.assertion import assert_or_throw
+from triad.utils.pandas_like import PD_UTILS
 
 
 class DataFrame(ABC):
@@ -19,14 +21,14 @@ class DataFrame(ABC):
     :param schema: |SchemaLikeObject|
     :param metadata: dict-like object with string keys, default ``None``
 
-    :Notice:
+    .. note::
 
-    This is an abstract class, and normally you don't construct it by yourself
-    unless you are
-    implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
+        This is an abstract class, and normally you don't construct it by yourself
+        unless you are
+        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
     """
 
-    _SHOW_LOCK = RLock()
+    _SHOW_LOCK = SerializableRLock()
 
     def __init__(self, schema: Any = None, metadata: Any = None):
         if not callable(schema):
@@ -43,7 +45,7 @@ class DataFrame(ABC):
             else ParamDict(metadata, deep=True)
         )
         self._metadata.set_readonly()
-        self._lazy_schema_lock = RLock()
+        self._lazy_schema_lock = SerializableRLock()
 
     @property
     def metadata(self) -> ParamDict:
@@ -127,7 +129,7 @@ class DataFrame(ABC):
     def as_pandas(self) -> pd.DataFrame:
         """Convert to pandas DataFrame"""
         pdf = pd.DataFrame(self.as_array(), columns=self.schema.names)
-        return _enforce_type(pdf, self.schema)
+        return PD_UTILS.enforce_type(pdf, self.schema.pa_schema, null_safe=True)
 
     def as_arrow(self, type_safe: bool = False) -> pa.Table:
         """Convert to pyArrow DataFrame"""
@@ -149,9 +151,9 @@ class DataFrame(ABC):
           defaults to False
         :return: 2-dimensional native python array
 
-        :Notice:
+        .. note::
 
-        If ``type_safe`` is False, then the returned values are 'raw' values.
+            If ``type_safe`` is False, then the returned values are 'raw' values.
         """
         raise NotImplementedError
 
@@ -166,9 +168,9 @@ class DataFrame(ABC):
           defaults to False
         :return: iterable of native python arrays
 
-        :Notice:
+        .. note::
 
-        If ``type_safe`` is False, then the returned values are 'raw' values.
+            If ``type_safe`` is False, then the returned values are 'raw' values.
         """
 
         raise NotImplementedError
@@ -186,6 +188,17 @@ class DataFrame(ABC):
         """
         raise NotImplementedError
 
+    @abstractmethod
+    def alter_columns(self, columns: Any) -> "DataFrame":  # pragma: no cover
+        """Change column types
+
+        :param columns: |SchemaLikeObject|,
+          all columns should be contained by the dataframe schema
+        :return: a new dataframe with altered columns, the order of the
+          original schema will not change
+        """
+        raise NotImplementedError
+
     @abstractmethod
     def _select_cols(self, cols: List[Any]) -> "DataFrame":  # pragma: no cover
         raise NotImplementedError
@@ -202,7 +215,7 @@ class DataFrame(ABC):
         try:
             schema = self.schema - columns
         except Exception as e:
-            raise FugueDataFrameOperationError(e)
+            raise FugueDataFrameOperationError from e
         if len(schema) == 0:
             raise FugueDataFrameOperationError(
                 "can't remove all columns of a dataframe"
@@ -219,7 +232,7 @@ class DataFrame(ABC):
         try:
             schema = self.schema.extract(columns)
         except Exception as e:
-            raise FugueDataFrameOperationError(e)
+            raise FugueDataFrameOperationError from e
         if len(schema) == 0:
             raise FugueDataFrameOperationError("must select at least one column")
         return self._select_cols(columns)
@@ -238,12 +251,12 @@ class DataFrame(ABC):
         :param title: title of the dataframe, defaults to None
         :param best_width: max width of the output table, defaults to 100
 
-        :Notice:
+        .. note::
 
-        When ``show_count`` is True, it can trigger expensive calculation for
-        a distributed dataframe. So if you call this function directly, you may
-        need to :func:`fugue.execution.execution_engine.ExecutionEngine.persist`
-        the dataframe.
+            When ``show_count`` is True, it can trigger expensive calculation for
+            a distributed dataframe. So if you call this function directly, you may
+            need to :func:`fugue.execution.execution_engine.ExecutionEngine.persist`
+            the dataframe.
         """
         self._show(
             head_rows=self.head(rows),
@@ -276,9 +289,9 @@ class DataFrame(ABC):
         :param columns: columns to extract, defaults to None
         :return: iterable of native python dicts
 
-        :Notice:
+        .. note::
 
-        The default implementation enforces ``type_safe`` True
+            The default implementation enforces ``type_safe`` True
         """
         if columns is None:
             columns = self.schema.names
@@ -337,25 +350,48 @@ class DataFrame(ABC):
                     print(self.metadata)
                 print("")
 
+    def _get_altered_schema(self, subschema: Any) -> Schema:
+        sub = Schema(subschema)
+        assert_or_throw(
+            sub.names in self.schema,
+            lambda: FugueDataFrameOperationError(
+                f"{sub.names} are not all in {self.schema}"
+            ),
+        )
+        for k, v in sub.items():
+            old_type = self.schema[k].type
+            new_type = v.type
+            if not old_type.equals(new_type):
+                assert_or_throw(
+                    not pa.types.is_struct(old_type)
+                    and not pa.types.is_list(old_type)
+                    and not pa.types.is_binary(old_type),
+                    lambda: NotImplementedError(f"can't convert from {old_type}"),
+                )
+                assert_or_throw(
+                    not pa.types.is_struct(new_type)
+                    and not pa.types.is_list(new_type)
+                    and not pa.types.is_binary(new_type),
+                    lambda: NotImplementedError(f"can't convert to {new_type}"),
+                )
+        return Schema([(k, sub.get(k, v)) for k, v in self.schema.items()])
+
 
 class LocalDataFrame(DataFrame):
     """Base class of all local dataframes. Please read
-    :ref:`this <tutorial:/tutorials/schema_dataframes.ipynb#dataframe>`
+    :ref:`this <tutorial:tutorials/advanced/schema_dataframes:dataframe>`
     to understand the concept
 
     :param schema: a `schema-like <triad.collections.schema.Schema>`_ object
     :param metadata: dict-like object with string keys, default ``None``
 
-    :Notice:
+    .. note::
 
-    This is an abstract class, and normally you don't construct it by yourself
-    unless you are
-    implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
+        This is an abstract class, and normally you don't construct it by yourself
+        unless you are
+        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
     """
 
-    def __init__(self, schema: Any = None, metadata: Any = None):
-        super().__init__(schema=schema, metadata=metadata)
-
     @property
     def is_local(self) -> bool:
         """Always True because it's a LocalDataFrame"""
@@ -373,22 +409,19 @@ class LocalDataFrame(DataFrame):
 
 class LocalBoundedDataFrame(LocalDataFrame):
     """Base class of all local bounded dataframes. Please read
-    :ref:`this <tutorial:/tutorials/schema_dataframes.ipynb#dataframe>`
+    :ref:`this <tutorial:tutorials/advanced/schema_dataframes:dataframe>`
     to understand the concept
 
     :param schema: |SchemaLikeObject|
     :param metadata: dict-like object with string keys, default ``None``
 
-    :Notice:
+    .. note::
 
-    This is an abstract class, and normally you don't construct it by yourself
-    unless you are
-    implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
+        This is an abstract class, and normally you don't construct it by yourself
+        unless you are
+        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
     """
 
-    def __init__(self, schema: Any = None, metadata: Any = None):
-        super().__init__(schema=schema, metadata=metadata)
-
     @property
     def is_bounded(self) -> bool:
         """Always True because it's a bounded dataframe"""
@@ -398,24 +431,22 @@ class LocalBoundedDataFrame(LocalDataFrame):
 class LocalUnboundedDataFrame(LocalDataFrame):
     """Base class of all local unbounded dataframes. Read
     this <https://fugue-tutorials.readthedocs.io/
-    en/latest/tutorials/schema_dataframes.html#DataFrame>`_ to understand the concept
+    en/latest/tutorials/advanced/schema_dataframes.html#DataFrame>`_
+    to understand the concept
 
     :param schema: |SchemaLikeObject|
     :param metadata: dict-like object with string keys, default ``None``
 
-    :Notice:
+    .. note::
 
-    This is an abstract class, and normally you don't construct it by yourself
-    unless you are
-    implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
+        This is an abstract class, and normally you don't construct it by yourself
+        unless you are
+        implementing a new :class:`~fugue.execution.execution_engine.ExecutionEngine`
     """
 
-    def __init__(self, schema: Any = None, metadata: Any = None):
-        super().__init__(schema=schema, metadata=metadata)
-
     @property
     def is_bounded(self):
-        """Always True because it's an unbounded dataframe"""
+        """Always False because it's an unbounded dataframe"""
         return False
 
     def count(self) -> int:
@@ -425,6 +456,38 @@ class LocalUnboundedDataFrame(LocalDataFrame):
         raise InvalidOperationError("Impossible to count an LocalUnboundedDataFrame")
 
 
+class YieldedDataFrame(Yielded):
+    """Yielded dataframe from :class:`~fugue.workflow.workflow.FugueWorkflow`.
+    Users shouldn't create this object directly.
+
+    :param yid: unique id for determinism
+    """
+
+    def __init__(self, yid: str):
+        super().__init__(yid)
+        self._df: Any = None
+
+    @property
+    def is_set(self) -> bool:
+        return self._df is not None
+
+    def set_value(self, df: DataFrame) -> None:
+        """Set the yielded dataframe after compute. Users should not
+        call it.
+
+        :param path: file path
+        """
+        self._df = df
+
+    @property
+    def result(self) -> DataFrame:
+        """The yielded dataframe, it will be set after the parent
+        workflow is computed
+        """
+        assert_or_throw(self.is_set, "value is not set")
+        return self._df
+
+
 class _PrettyTable(object):
     def __init__(
         self,  # noqa: C901
@@ -536,21 +599,3 @@ def _get_schema_change(
 
 def _input_schema(schema: Any) -> Schema:
     return schema if isinstance(schema, Schema) else Schema(schema)
-
-
-def _enforce_type(df: pd.DataFrame, schema: Schema) -> pd.DataFrame:
-    # TODO: does this have higher latency?
-    for k, v in schema.items():
-        s = df[k]
-        if pa.types.is_string(v.type):
-            ns = s.isnull()
-            s = s.astype(str)
-            s[ns] = None
-        elif pa.types.is_integer(v.type) or pa.types.is_boolean(v.type):
-            ns = s.isnull()
-            s = s.fillna(0).astype(v.type.to_pandas_dtype())
-            s[ns] = None
-        elif not pa.types.is_struct(v.type) and not pa.types.is_list(v.type):
-            s = s.astype(v.type.to_pandas_dtype())
-        df[k] = s
-    return df
