@@ -5,10 +5,10 @@ from fugue.dataframe.dataframe import (
     LocalUnboundedDataFrame,
     _get_schema_change,
 )
+from fugue.exceptions import FugueDataFrameOperationError
 from triad.collections.schema import Schema
 from triad.utils.iter import EmptyAwareIterable, make_empty_aware
 from triad.utils.pyarrow import apply_schema
-from fugue.exceptions import FugueDataFrameInitError, FugueDataFrameOperationError
 
 
 class IterableDataFrame(LocalUnboundedDataFrame):
@@ -20,43 +20,38 @@ class IterableDataFrame(LocalUnboundedDataFrame):
     :param schema: |SchemaLikeObject|
     :param metadata: dict-like object with string keys, default ``None``
 
-    :raises FugueDataFrameInitError: if the input is not compatible
+    .. admonition:: Examples
 
-    :Examples:
+        >>> a = IterableDataFrame([[0,'a'],[1,'b']],"a:int,b:str")
+        >>> b = IterableDataFrame(a)
 
-    >>> a = IterableDataFrame([[0,'a'],[1,'b']],"a:int,b:str")
-    >>> b = IterableDataFrame(a)
+    .. note::
 
-    :Notice:
-
-    It's ok to peek the dataframe, it will not affect the iteration, but it's
-    invalid operation to count
+        It's ok to peek the dataframe, it will not affect the iteration, but it's
+        invalid operation to count
     """
 
     def __init__(  # noqa: C901
         self, df: Any = None, schema: Any = None, metadata: Any = None
     ):
-        try:
-            if df is None:
-                idf: Iterable[Any] = []
-                orig_schema: Optional[Schema] = None
-            elif isinstance(df, IterableDataFrame):
-                idf = df.native
-                orig_schema = df.schema
-            elif isinstance(df, DataFrame):
-                idf = df.as_array_iterable(type_safe=False)
-                orig_schema = df.schema
-            elif isinstance(df, Iterable):
-                idf = df
-                orig_schema = None
-            else:
-                raise ValueError(f"{df} is incompatible with IterableDataFrame")
-            schema, pos = _get_schema_change(orig_schema, schema)
-            super().__init__(schema, metadata)
-            self._pos = pos
-            self._native = make_empty_aware(self._preprocess(idf))
-        except Exception as e:
-            raise FugueDataFrameInitError(e)
+        if df is None:
+            idf: Iterable[Any] = []
+            orig_schema: Optional[Schema] = None
+        elif isinstance(df, IterableDataFrame):
+            idf = df.native
+            orig_schema = df.schema
+        elif isinstance(df, DataFrame):
+            idf = df.as_array_iterable(type_safe=False)
+            orig_schema = df.schema
+        elif isinstance(df, Iterable):
+            idf = df
+            orig_schema = None
+        else:
+            raise ValueError(f"{df} is incompatible with IterableDataFrame")
+        schema, pos = _get_schema_change(orig_schema, schema)
+        super().__init__(schema, metadata)
+        self._pos = pos
+        self._native = make_empty_aware(self._preprocess(idf))
 
     @property
     def native(self) -> EmptyAwareIterable[Any]:
@@ -77,13 +72,19 @@ class IterableDataFrame(LocalUnboundedDataFrame):
     def _select_cols(self, keys: List[Any]) -> DataFrame:
         return IterableDataFrame(self, self.schema.extract(keys))
 
-    def rename(self, columns: Dict[str, str]) -> "DataFrame":
+    def rename(self, columns: Dict[str, str]) -> DataFrame:
         try:
             schema = self.schema.rename(columns)
         except Exception as e:
-            raise FugueDataFrameOperationError(e)
+            raise FugueDataFrameOperationError from e
         return IterableDataFrame(self.native, schema)
 
+    def alter_columns(self, columns: Any) -> DataFrame:
+        new_schema = self._get_altered_schema(columns)
+        if new_schema == self.schema:
+            return self
+        return IterableDataFrame(self.native, new_schema)
+
     def as_array(
         self, columns: Optional[List[str]] = None, type_safe: bool = False
     ) -> List[Any]:
