@@ -1,22 +1,24 @@
 import logging
 from abc import ABC, abstractmethod
 from typing import Any, Callable, Dict, Iterable, List, Optional, Union
+from uuid import uuid4
 
 from fugue.collections.partition import (
     EMPTY_PARTITION_SPEC,
     PartitionCursor,
     PartitionSpec,
 )
-from fugue.constants import FUGUE_DEFAULT_CONF
+from fugue.column import ColumnExpr, SelectColumns, SQLExpressionGenerator, col, is_agg
+from fugue.constants import _FUGUE_GLOBAL_CONF
 from fugue.dataframe import DataFrame, DataFrames
 from fugue.dataframe.array_dataframe import ArrayDataFrame
 from fugue.dataframe.dataframe import LocalDataFrame
 from fugue.dataframe.utils import deserialize_df, serialize_df
 from fugue.exceptions import FugueBug
-from triad.collections import ParamDict, Schema
+from fugue.rpc import RPCServer, make_rpc_server
+from triad import ParamDict, Schema, SerializableRLock, assert_or_throw
 from triad.collections.fs import FileSystem
 from triad.exceptions import InvalidOperationError
-from triad.utils.assertion import assert_or_throw
 from triad.utils.convert import to_size
 from triad.utils.string import validate_triad_var_name
 
@@ -25,7 +27,7 @@ _DEFAULT_JOIN_KEYS: List[str] = []
 
 class SQLEngine(ABC):
     """The abstract base class for different SQL execution implementations. Please read
-    :ref:`this <tutorial:/tutorials/execution_engine.ipynb#sqlengine>`
+    :ref:`this <tutorial:tutorials/advanced/execution_engine:sqlengine>`
     to understand the concept
 
     :param execution_engine: the execution engine this sql engine will run on
@@ -47,17 +49,17 @@ class SQLEngine(ABC):
         :param statement: the ``SELECT`` statement using the ``dfs`` keys as tables
         :return: result of the ``SELECT`` statement
 
-        :Example:
+        .. admonition:: Examples
 
-        >>> dfs = DataFrames(a=df1, b=df2)
-        >>> sql_engine.select(dfs, "SELECT * FROM a UNION SELECT * FROM b")
+            >>> dfs = DataFrames(a=df1, b=df2)
+            >>> sql_engine.select(dfs, "SELECT * FROM a UNION SELECT * FROM b")
 
-        :Notice:
+        .. note::
 
-        There can be tables that is not in ``dfs``. For example you want to select
-        from hive without input DataFrames:
+            There can be tables that is not in ``dfs``. For example you want to select
+            from hive without input DataFrames:
 
-        >>> sql_engine.select(DataFrames(), "SELECT * FROM hive.a.table")
+            >>> sql_engine.select(DataFrames(), "SELECT * FROM hive.a.table")
         """
         raise NotImplementedError
 
@@ -65,34 +67,106 @@ class SQLEngine(ABC):
 class ExecutionEngine(ABC):
     """The abstract base class for execution engines.
     It is the layer that unifies core concepts of distributed computing,
-    and separates the underlying computing frameworks from user’s higher level logic.
+    and separates the underlying computing frameworks from user's higher level logic.
 
-    Please read
-    :ref:`The ExecutionEngine Tutorial <tutorial:/tutorials/execution_engine.ipynb>`
+    Please read |ExecutionEngineTutorial|
     to understand this most important Fugue concept
 
     :param conf: dict-like config, read
-      :ref:`this <tutorial:/tutorials/useful_config.ipynb>`
+      :doc:`this <tutorial:tutorials/advanced/useful_config>`
       to learn Fugue specific options
     """
 
     def __init__(self, conf: Any):
         _conf = ParamDict(conf)
-        self._conf = ParamDict({**FUGUE_DEFAULT_CONF, **_conf})
+        self._conf = ParamDict({**_FUGUE_GLOBAL_CONF, **_conf})
+        self._compile_conf = ParamDict()
+        self._rpc_server = make_rpc_server(self.conf)
+        self._engine_start_lock = SerializableRLock()
+        self._engine_start = 0
+        self._sql_engine: Optional[SQLEngine] = None
+
+    def start(self) -> None:
+        """Start this execution engine, do not override.
+        You should customize :meth:`~.start_engine` if necessary.
+        """
+        with self._engine_start_lock:
+            if self._engine_start == 0:
+                self.rpc_server.start()
+                self.start_engine()
+            self._engine_start += 1
+
+    def stop(self) -> None:
+        """Stop this execution engine, do not override
+        You should customize :meth:`~.stop_engine` if necessary.
+        """
+        with self._engine_start_lock:
+            if self._engine_start == 1:
+                try:
+                    self.stop_engine()
+                finally:
+                    self._engine_start -= 1
+                    self.rpc_server.stop()
+            else:
+                self._engine_start -= 1
+            self._engine_start = max(0, self._engine_start)
+
+    def start_engine(self) -> None:  # pragma: no cover
+        """Custom logic to start the execution engine, defaults to no operation"""
+        return
+
+    def stop_engine(self) -> None:  # pragma: no cover
+        """Custom logic to stop the execution engine, defaults to no operation"""
+        return
 
     @property
     def conf(self) -> ParamDict:
         """All configurations of this engine instance.
 
-        :Notice:
+        .. note::
 
-        it can contain more than you providec, for example
-        in :class:`~fugue_spark.execution_engine.SparkExecutionEngine`,
-        the Spark session can bring in more config, they are all accessible
-        using this property.
+            It can contain more than you providec, for example
+            in :class:`~fugue_spark.execution_engine.SparkExecutionEngine`,
+            the Spark session can bring in more config, they are all accessible
+            using this property.
         """
         return self._conf
 
+    @property
+    def compile_conf(self) -> ParamDict:
+        """Compiled time (workflow level) configurations, it is always a
+        superset of conf
+
+        .. note::
+
+            Users normally don't need to use this property. It is for
+            internal use.
+        """
+        return self._compile_conf
+
+    @property
+    def rpc_server(self) -> RPCServer:
+        """:class:`~fugue.rpc.base.RPCServer` of this execution engine"""
+        return self._rpc_server
+
+    @property
+    def sql_engine(self) -> SQLEngine:
+        """The :class:`~.SQLEngine` currently used by this execution engine.
+        You should use :meth:`~.set_sql_engine` to set a new SQLEngine
+        instance. If not set, the default is :meth:`~.default_sql_engine`
+        """
+        if self._sql_engine is None:
+            self._sql_engine = self.default_sql_engine
+        return self._sql_engine
+
+    def set_sql_engine(self, engine: SQLEngine) -> None:
+        """Set :class:`~.SQLEngine` for this execution engine.
+        If not set, the default is :meth:`~.default_sql_engine`
+
+        :param engine: :class:`~.SQLEngine` instance
+        """
+        self._sql_engine = engine
+
     @property
     @abstractmethod
     def log(self) -> logging.Logger:  # pragma: no cover
@@ -111,11 +185,6 @@ class ExecutionEngine(ABC):
         """Default SQLEngine if user doesn't specify"""
         raise NotImplementedError
 
-    @abstractmethod
-    def stop(self) -> None:  # pragma: no cover
-        """Stop this execution engine"""
-        raise NotImplementedError
-
     @abstractmethod
     def to_df(
         self, data: Any, schema: Any = None, metadata: Any = None
@@ -129,13 +198,14 @@ class ExecutionEngine(ABC):
         :param metadata: |ParamsLikeObject|, defaults to None
         :return: engine compatible dataframe
 
-        :Notice:
+        .. note::
 
-        There are certain conventions to follow for a new implementation:
+            There are certain conventions to follow for a new implementation:
 
-        * if the input is already in compatible dataframe type, it should return itself
-        * all other methods in the engine interface should take arbitrary dataframes and
-          call this method to convert before doing anything
+            * if the input is already in compatible dataframe type, it should return
+              itself
+            * all other methods in the engine interface should take arbitrary
+              dataframes and call this method to convert before doing anything
         """
         raise NotImplementedError
 
@@ -149,9 +219,9 @@ class ExecutionEngine(ABC):
         :param partition_spec: how you want to partition the dataframe
         :return: repartitioned dataframe
 
-        :Notice:
+        .. note::
 
-        Before implementing please read |PartitionTutorial|
+            Before implementing please read |PartitionTutorial|
         """
         raise NotImplementedError
 
@@ -172,7 +242,7 @@ class ExecutionEngine(ABC):
         :param map_func: the function to apply on every logical partition
         :param output_schema: |SchemaLikeObject| that can't be None.
           Please also understand :ref:`why we need this
-          <tutorial:/tutorials/transformer.ipynb#why-explicit-on-output-schema?>`
+          <tutorial:tutorials/beginner/interface:schema>`
         :param partition_spec: partition specification
         :param metadata: dict-like metadata object to add to the dataframe after the
           map operation, defaults to None
@@ -180,11 +250,11 @@ class ExecutionEngine(ABC):
           defaults to None
         :return: the dataframe after the map operation
 
-        :Notice:
+        .. note::
 
-        Before implementing, you must read
-        :ref:`this <tutorial:/tutorials/execution_engine.ipynb#map>` to understand
-        what map is used for and how it should work.
+            Before implementing, you must read
+            :ref:`this <tutorial:tutorials/advanced/execution_engine:map>`
+            to understand what map is used for and how it should work.
         """
         raise NotImplementedError
 
@@ -210,17 +280,17 @@ class ExecutionEngine(ABC):
         :param lazy: ``True``: first usage of the output will trigger persisting
           to happen; ``False`` (eager): persist is forced to happend immediately.
           Default to ``False``
-        :param *args: parameter to pass to the underlying persist implementation
-        :param *kwargs: parameter to pass to the underlying persist implementation
+        :param args: parameter to pass to the underlying persist implementation
+        :param kwargs: parameter to pass to the underlying persist implementation
         :return: the persisted dataframe
 
-        :Notice:
+        .. note::
 
-        ``persist`` can only guarantee the persisted dataframe will be computed
-        for only once. However this doesn't mean the backend really breaks up the
-        execution dependency at the persisting point. Commonly, it doesn't cause
-        any issue, but if your execution graph is long, it may cause expected
-        problems for example, stack overflow.
+            ``persist`` can only guarantee the persisted dataframe will be computed
+            for only once. However this doesn't mean the backend really breaks up the
+            execution dependency at the persisting point. Commonly, it doesn't cause
+            any issue, but if your execution graph is long, it may cause expected
+            problems for example, stack overflow.
         """
         raise NotImplementedError
 
@@ -245,9 +315,9 @@ class ExecutionEngine(ABC):
           defaults to None
         :return: the joined dataframe
 
-        :Notice:
+        .. note::
 
-        Please read :func:`this <fugue.dataframe.utils.get_join_schemas>`
+            Please read :func:`this <fugue.dataframe.utils.get_join_schemas>`
         """
         raise NotImplementedError
 
@@ -269,10 +339,10 @@ class ExecutionEngine(ABC):
           defaults to None
         :return: the unioned dataframe
 
-        :Notice:
+        .. note::
 
-        Currently, the schema of ``df1`` and ``df2`` must be identical, or
-        an exception will be thrown.
+            Currently, the schema of ``df1`` and ``df2`` must be identical, or
+            an exception will be thrown.
         """
         raise NotImplementedError
 
@@ -294,10 +364,10 @@ class ExecutionEngine(ABC):
           defaults to None
         :return: the unioned dataframe
 
-        :Notice:
+        .. note::
 
-        Currently, the schema of ``df1`` and ``df2`` must be identical, or
-        an exception will be thrown.
+            Currently, the schema of ``df1`` and ``df2`` must be identical, or
+            an exception will be thrown.
         """
         raise NotImplementedError
 
@@ -319,10 +389,10 @@ class ExecutionEngine(ABC):
           defaults to None
         :return: the unioned dataframe
 
-        :Notice:
+        .. note::
 
-        Currently, the schema of ``df1`` and ``df2`` must be identical, or
-        an exception will be thrown.
+            Currently, the schema of ``df1`` and ``df2`` must be identical, or
+            an exception will be thrown.
         """
         raise NotImplementedError
 
@@ -347,27 +417,347 @@ class ExecutionEngine(ABC):
     def dropna(
         self,
         df: DataFrame,
-        metadata: Any = None,
         how: str = "any",
         thresh: int = None,
         subset: List[str] = None,
+        metadata: Any = None,
     ) -> DataFrame:  # pragma: no cover
         """Drop NA recods from dataframe
 
         :param df: DataFrame
-        :param metadata: dict-like object to add to the result dataframe,
-            defaults to None
-        :type metadata: Any, optional
         :param how: 'any' or 'all'. 'any' drops rows that contain any nulls.
           'all' drops rows that contain all nulls.
         :param thresh: int, drops rows that have less than thresh non-null values
         :param subset: list of columns to operate on
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None
+        :type metadata: Any, optional
 
         :return: DataFrame with NA records dropped
         :rtype: DataFrame
         """
         pass
 
+    @abstractmethod
+    def fillna(
+        self, df: DataFrame, value: Any, subset: List[str] = None, metadata: Any = None
+    ) -> DataFrame:  # pragma: no cover
+        """
+        Fill ``NULL``, ``NAN``, ``NAT`` values in a dataframe
+
+        :param df: DataFrame
+        :param value: if scalar, fills all columns with same value.
+            if dictionary, fills NA using the keys as column names and the
+            values as the replacement values.
+        :param subset: list of columns to operate on. ignored if value is
+            a dictionary
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None
+        :type metadata: Any, optional
+
+        :return: DataFrame with NA records filled
+        :rtype: DataFrame
+        """
+        pass
+
+    @abstractmethod
+    def sample(
+        self,
+        df: DataFrame,
+        n: Optional[int] = None,
+        frac: Optional[float] = None,
+        replace: bool = False,
+        seed: Optional[int] = None,
+        metadata: Any = None,
+    ) -> DataFrame:  # pragma: no cover
+        """
+        Sample dataframe by number of rows or by fraction
+
+        :param df: DataFrame
+        :param n: number of rows to sample, one and only one of ``n`` and ``frac``
+          must be set
+        :param frac: fraction [0,1] to sample, one and only one of ``n`` and ``frac``
+          must be set
+        :param replace: whether replacement is allowed. With replacement,
+          there may be duplicated rows in the result, defaults to False
+        :param seed: seed for randomness, defaults to None
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None
+
+        :return: sampled dataframe
+        :rtype: DataFrame
+        """
+        pass
+
+    @abstractmethod
+    def take(
+        self,
+        df: DataFrame,
+        n: int,
+        presort: str,
+        na_position: str = "last",
+        partition_spec: PartitionSpec = EMPTY_PARTITION_SPEC,
+        metadata: Any = None,
+    ) -> DataFrame:  # pragma: no cover
+        """
+        Get the first n rows of a DataFrame per partition. If a presort is defined,
+        use the presort before applying take. presort overrides partition_spec.presort.
+        The Fugue implementation of the presort follows Pandas convention of specifying
+        NULLs first or NULLs last. This is different from the Spark and SQL convention
+        of NULLs as the smallest value.
+
+        :param df: DataFrame
+        :param n: number of rows to return
+        :param presort: presort expression similar to partition presort
+        :param na_position: position of null values during the presort.
+            can accept ``first`` or ``last``
+        :param partition_spec: PartitionSpec to apply the take operation
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None
+
+        :return: n rows of DataFrame per partition
+        :rtype: DataFrame
+        """
+        pass
+
+    def select(
+        self,
+        df: DataFrame,
+        cols: SelectColumns,
+        where: Optional[ColumnExpr] = None,
+        having: Optional[ColumnExpr] = None,
+        metadata: Any = None,
+    ) -> DataFrame:
+        """The functional interface for SQL select statement
+
+        :param df: the dataframe to be operated on
+        :param cols: column expressions
+        :param where: ``WHERE`` condition expression, defaults to None
+        :param having: ``having`` condition expression, defaults to None. It
+          is used when ``cols`` contains aggregation columns, defaults to None
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None. It's currently not used
+        :return: the select result as a dataframe
+
+        .. admonition:: New Since
+            :class: hint
+
+            **0.6.0**
+
+        .. attention::
+
+            This interface is experimental, it's subjected to change in new versions.
+
+        .. seealso::
+
+            Please find more expression examples in :mod:`fugue.column.sql` and
+            :mod:`fugue.column.functions`
+
+        .. admonition:: Examples
+
+            .. code-block:: python
+
+                import fugue.column.functions as f
+
+                # select existed and new columns
+                engine.select(df, SelectColumns(col("a"),col("b"),lit(1,"another")))
+                engine.select(df, SelectColumns(col("a"),(col("b")+lit(1)).alias("x")))
+
+                # aggregation
+                # SELECT COUNT(DISTINCT *) AS x FROM df
+                engine.select(
+                    df,
+                    SelectColumns(f.count_distinct(col("*")).alias("x")))
+
+                # SELECT a, MAX(b+1) AS x FROM df GROUP BY a
+                engine.select(
+                    df,
+                    SelectColumns(col("a"),f.max(col("b")+lit(1)).alias("x")))
+
+                # SELECT a, MAX(b+1) AS x FROM df
+                #   WHERE b<2 AND a>1
+                #   GROUP BY a
+                #   HAVING MAX(b+1)>0
+                engine.select(
+                    df,
+                    SelectColumns(col("a"),f.max(col("b")+lit(1)).alias("x")),
+                    where=(col("b")<2) & (col("a")>1),
+                    having=f.max(col("b")+lit(1))>0
+                )
+        """
+        gen = SQLExpressionGenerator(enable_cast=False)
+        df_name = _get_temp_df_name()
+        sql = gen.select(cols, df_name, where=where, having=having)
+        res = self.sql_engine.select(DataFrames({df_name: self.to_df(df)}), sql)
+        diff = gen.correct_select_schema(df.schema, cols, res.schema)
+        return res if diff is None else res.alter_columns(diff)
+
+    def filter(
+        self, df: DataFrame, condition: ColumnExpr, metadata: Any = None
+    ) -> DataFrame:
+        """Filter rows by the given condition
+
+        :param df: the dataframe to be filtered
+        :param condition: (boolean) column expression
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None. It's currently not used
+        :return: the filtered dataframe
+
+        .. admonition:: New Since
+            :class: hint
+
+            **0.6.0**
+
+        .. seealso::
+
+            Please find more expression examples in :mod:`fugue.column.sql` and
+            :mod:`fugue.column.functions`
+
+        .. admonition:: Examples
+
+            .. code-block:: python
+
+                import fugue.column.functions as f
+
+                engine.filter(df, (col("a")>1) & (col("b")=="x"))
+                engine.filter(df, f.coalesce(col("a"),col("b"))>1)
+        """
+        return self.select(
+            df, cols=SelectColumns(col("*")), where=condition, metadata=metadata
+        )
+
+    def assign(
+        self, df: DataFrame, columns: List[ColumnExpr], metadata: Any = None
+    ) -> DataFrame:
+        """Update existing columns with new values and add new columns
+
+        :param df: the dataframe to set columns
+        :param columns: column expressions
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None. It's currently not used
+        :return: the updated dataframe
+
+        .. tip::
+
+            This can be used to cast data types, alter column values or add new
+            columns. But you can't use aggregation in columns.
+
+        .. admonition:: New Since
+            :class: hint
+
+            **0.6.0**
+
+        .. seealso::
+
+            Please find more expression examples in :mod:`fugue.column.sql` and
+            :mod:`fugue.column.functions`
+
+        .. admonition:: Examples
+
+            .. code-block:: python
+
+                # assume df has schema: a:int,b:str
+
+                # add constant column x
+                engine.assign(df, lit(1,"x"))
+
+                # change column b to be a constant integer
+                engine.assign(df, lit(1,"b"))
+
+                # add new x to be a+b
+                engine.assign(df, (col("a")+col("b")).alias("x"))
+
+                # cast column a data type to double
+                engine.assign(df, col("a").cast(float))
+        """
+        SelectColumns(
+            *columns
+        ).assert_no_wildcard().assert_all_with_names().assert_no_agg()
+
+        cols = [col(n) for n in df.schema.names]
+        for c in columns:
+            if c.output_name not in df.schema:
+                cols.append(c)
+            else:
+                cols[df.schema.index_of_key(c.output_name)] = c
+        return self.select(df, SelectColumns(*cols), metadata=metadata)
+
+    def aggregate(
+        self,
+        df: DataFrame,
+        partition_spec: Optional[PartitionSpec],
+        agg_cols: List[ColumnExpr],
+        metadata: Any = None,
+    ):
+        """Aggregate on dataframe
+
+        :param df: the dataframe to aggregate on
+        :param partition_spec: PartitionSpec to specify partition keys
+        :param agg_cols: aggregation expressions
+        :param metadata: dict-like object to add to the result dataframe,
+            defaults to None. It's currently not used
+        :return: the aggregated result as a dataframe
+
+        .. admonition:: New Since
+            :class: hint
+
+            **0.6.0**
+
+        .. seealso::
+
+            Please find more expression examples in :mod:`fugue.column.sql` and
+            :mod:`fugue.column.functions`
+
+        .. admonition:: Examples
+
+            .. code-block:: python
+
+                import fugue.column.functions as f
+
+                # SELECT MAX(b) AS b FROM df
+                engine.aggregate(
+                    df,
+                    partition_spec=None,
+                    agg_cols=[f.max(col("b"))])
+
+                # SELECT a, MAX(b) AS x FROM df GROUP BY a
+                engine.aggregate(
+                    df,
+                    partition_spec=PartitionSpec(by=["a"]),
+                    agg_cols=[f.max(col("b")).alias("x")])
+        """
+        assert_or_throw(len(agg_cols) > 0, ValueError("agg_cols can't be empty"))
+        assert_or_throw(
+            all(is_agg(x) for x in agg_cols),
+            ValueError("all agg_cols must be aggregation functions"),
+        )
+        keys: List[ColumnExpr] = []
+        if partition_spec is not None and len(partition_spec.partition_by) > 0:
+            keys = [col(y) for y in partition_spec.partition_by]
+        cols = SelectColumns(*keys, *agg_cols)
+        return self.select(df, cols=cols, metadata=metadata)
+
+    def convert_yield_dataframe(self, df: DataFrame, as_local: bool) -> DataFrame:
+        """Convert a yield dataframe to a dataframe that can be used after this
+        execution engine stops.
+
+        :param df: DataFrame
+        :param as_local: whether yield a local dataframe
+        :return: another DataFrame that can be used after this execution engine stops
+
+        .. note::
+
+            By default, the output dataframe is the input dataframe. But it should be
+            overridden if when an engine stops and the input dataframe will become
+            invalid.
+
+            For example, if you custom a spark engine where you start and stop the spark
+            session in this engine's :meth:`~.start_engine` and :meth:`~.stop_engine`,
+            then the spark dataframe will be invalid. So you may consider converting
+            it to a local dataframe so it can still exist after the engine stops.
+        """
+        return df.as_local() if as_local else df
+
     def zip(
         self,
         df1: DataFrame,
@@ -398,17 +788,18 @@ class ExecutionEngine(ABC):
         :return: a zipped dataframe, the metadata of the
           dataframe will indicate it's zipped
 
-        :Notice:
+        .. note::
 
-        * Different from join, ``df1`` and ``df2`` can have common columns that you will
-          not use as partition keys.
-        * If ``on`` is not specified it will also use the common columns of the two
-          dataframes (if it's not a cross zip)
-        * For non-cross zip, the two dataframes must have common columns, or error will
-          be thrown
+            * Different from join, ``df1`` and ``df2`` can have common columns that you
+              will not use as partition keys.
+            * If ``on`` is not specified it will also use the common columns of the two
+              dataframes (if it's not a cross zip)
+            * For non-cross zip, the two dataframes must have common columns, or error
+              will be thrown
 
-        For more details and examples, read
-        :ref:`Zip & Comap <tutorial:/tutorials/execution_engine.ipynb#zip-&-comap>`.
+        .. seealso::
+
+            For more details and examples, read |ZipComap|.
         """
         on = list(partition_spec.partition_by)
         how = how.lower()
@@ -444,7 +835,7 @@ class ExecutionEngine(ABC):
                 )
             for k in df.metadata["serialized_cols"].keys():
                 assert_or_throw(
-                    k not in serialized_cols, ValueError(f"{k} is duplicated")
+                    k not in serialized_cols, lambda: ValueError(f"{k} is duplicated")
                 )
                 serialized_cols[k] = df.metadata["serialized_cols"][k]
                 schemas[k] = df.metadata["schemas"][k]
@@ -482,15 +873,16 @@ class ExecutionEngine(ABC):
         :return: a zipped dataframe, the metadata of the
           dataframe will indicated it's zipped
 
-        :Notice:
+        .. note::
 
-        * Please also read :meth:`~.zip`
-        * If ``dfs`` is dict like, the zipped dataframe will be dict like,
-          If ``dfs`` is list like, the zipped dataframe will be list like
-        * It's fine to contain only one dataframe in ``dfs``
+            * Please also read :meth:`~.zip`
+            * If ``dfs`` is dict like, the zipped dataframe will be dict like,
+              If ``dfs`` is list like, the zipped dataframe will be list like
+            * It's fine to contain only one dataframe in ``dfs``
 
-        For more details and examples, read
-        :ref:`Zip & Comap <tutorial:/tutorials/execution_engine.ipynb#zip-&-comap>`.
+        .. seealso::
+
+            For more details and examples, read |ZipComap|
         """
         assert_or_throw(len(dfs) > 0, "can't zip 0 dataframes")
         pairs = list(dfs.items())
@@ -542,7 +934,7 @@ class ExecutionEngine(ABC):
         :param map_func: the function to apply on every zipped partition
         :param output_schema: |SchemaLikeObject| that can't be None.
           Please also understand :ref:`why we need this
-          <tutorial:/tutorials/cotransformer.ipynb#why-explicit-on-output-schema?>`
+          <tutorial:tutorials/beginner/interface:schema>`
         :param partition_spec: partition specification for processing the zipped
           zipped dataframe.
         :param metadata: dict-like metadata object to add to the dataframe after the
@@ -551,24 +943,25 @@ class ExecutionEngine(ABC):
           defaults to None
         :return: the dataframe after the comap operation
 
-        :Notice:
+        .. note::
 
-        * The input of this method must be an output of :meth:`~.zip` or
-          :meth:`~.zip_all`
-        * The ``partition_spec`` here is NOT related with how you zipped the dataframe
-          and however you set it, will only affect the processing speed, actually the
-          partition keys will be overriden to the zipped dataframe partition keys.
-          You may use it in this way to improve the efficiency:
-          ``PartitionSpec(algo="even", num="ROWCOUNT")``,
-          this tells the execution engine to put each zipped partition into a physical
-          partition so it can achieve the best possible load balance.
-        * If input dataframe has keys, the dataframes you get in ``map_func`` and
-          ``on_init`` will have keys, otherwise you will get list-like dataframes
-        * on_init function will get a DataFrames object that has the same structure, but
-          has all empty dataframes, you can use the schemas but not the data.
+            * The input of this method must be an output of :meth:`~.zip` or
+              :meth:`~.zip_all`
+            * The ``partition_spec`` here is NOT related with how you zipped the
+              dataframe and however you set it, will only affect the processing speed,
+              actually the partition keys will be overriden to the zipped dataframe
+              partition keys. You may use it in this way to improve the efficiency:
+              ``PartitionSpec(algo="even", num="ROWCOUNT")``,
+              this tells the execution engine to put each zipped partition into a
+              physical partition so it can achieve the best possible load balance.
+            * If input dataframe has keys, the dataframes you get in ``map_func`` and
+              ``on_init`` will have keys, otherwise you will get list-like dataframes
+            * on_init function will get a DataFrames object that has the same structure,
+              but has all empty dataframes, you can use the schemas but not the data.
 
-        For more details and examples, read
-        :ref:`Zip & Comap <tutorial:/tutorials/execution_engine.ipynb#zip-&-comap>`.
+        .. seealso::
+
+            For more details and examples, read |ZipComap|
         """
         assert_or_throw(df.metadata["serialized"], ValueError("df is not serilaized"))
         cs = _Comap(df, map_func, on_init)
@@ -595,8 +988,7 @@ class ExecutionEngine(ABC):
         :param kwargs: parameters to pass to the underlying framework
         :return: an engine compatible dataframe
 
-        For more details and examples, read
-        :ref:`Load & Save <tutorial:/tutorials/execution_engine.ipynb#load-&-save>`.
+        For more details and examples, read |ZipComap|.
         """
         raise NotImplementedError
 
@@ -624,8 +1016,7 @@ class ExecutionEngine(ABC):
         :param force_single: force the output as a single file, defaults to False
         :param kwargs: parameters to pass to the underlying framework
 
-        For more details and examples, read
-        :ref:`Load & Save <tutorial:/tutorials/execution_engine.ipynb#load-&-save>`.
+        For more details and examples, read |LoadSave|.
         """
         raise NotImplementedError
 
@@ -747,3 +1138,7 @@ def _generate_comap_empty_dfs(schemas: Any, named: bool) -> DataFrames:
         return DataFrames({k: ArrayDataFrame([], v) for k, v in schemas.items()})
     else:
         return DataFrames([ArrayDataFrame([], v) for v in schemas.values()])
+
+
+def _get_temp_df_name() -> str:
+    return "_" + str(uuid4())[:5]
