@@ -1,13 +1,22 @@
+import inspect
 import logging
-from typing import Any, Callable, List, Optional, Union
+import os
+from typing import Any, Callable, Dict, List, Optional, Union
 
 import pandas as pd
+from fugue._utils.interfaceless import (
+    ExecutionEngineParam,
+    SimpleAnnotationConverter,
+    register_annotation_converter,
+)
 from fugue._utils.io import load_df, save_df
 from fugue.collections.partition import (
     EMPTY_PARTITION_SPEC,
     PartitionCursor,
     PartitionSpec,
+    parse_presort_exp,
 )
+from fugue.constants import FUGUE_CONF_SQL_IGNORE_CASE
 from fugue.dataframe import (
     DataFrame,
     DataFrames,
@@ -22,10 +31,11 @@ from fugue.execution.execution_engine import (
     ExecutionEngine,
     SQLEngine,
 )
+from qpd_pandas import run_sql_on_pandas
 from qpd_pandas.engine import PandasUtils
 from sqlalchemy import create_engine
 from triad.collections import Schema
-from triad.collections.dict import ParamDict
+from triad.collections.dict import IndexedOrderedDict, ParamDict
 from triad.collections.fs import FileSystem
 from triad.utils.assertion import assert_or_throw
 
@@ -47,6 +57,30 @@ class SqliteEngine(SQLEngine):
         return PandasDataFrame(df)
 
 
+class QPDPandasEngine(SQLEngine):
+    """QPD execution implementation.
+
+    :param execution_engine: the execution engine this sql engine will run on
+    """
+
+    def __init__(self, execution_engine: ExecutionEngine):
+        super().__init__(execution_engine)
+
+    def select(self, dfs: DataFrames, statement: str) -> DataFrame:
+        _dfs = {
+            k: self.execution_engine.to_df(v).as_pandas()  # type: ignore
+            for k, v in dfs.items()
+        }
+        df = run_sql_on_pandas(
+            statement,
+            _dfs,
+            ignore_case=self.execution_engine.compile_conf.get(
+                FUGUE_CONF_SQL_IGNORE_CASE, False
+            ),
+        )
+        return self.execution_engine.to_df(df)
+
+
 class NativeExecutionEngine(ExecutionEngine):
     """The execution engine based on native python and pandas. This execution engine
     is mainly for prototyping and unit tests.
@@ -60,7 +94,6 @@ class NativeExecutionEngine(ExecutionEngine):
         super().__init__(conf)
         self._fs = FileSystem()
         self._log = logging.getLogger()
-        self._default_sql_engine = SqliteEngine(self)
 
     def __repr__(self) -> str:
         return "NativeExecutionEngine"
@@ -75,16 +108,13 @@ class NativeExecutionEngine(ExecutionEngine):
 
     @property
     def default_sql_engine(self) -> SQLEngine:
-        return self._default_sql_engine
+        return QPDPandasEngine(self)
 
     @property
     def pl_utils(self) -> PandasUtils:
         """Pandas-like dataframe utils"""
         return PandasUtils()
 
-    def stop(self) -> None:  # pragma: no cover
-        return
-
     def to_df(
         self, df: Any, schema: Any = None, metadata: Any = None
     ) -> LocalBoundedDataFrame:
@@ -93,7 +123,7 @@ class NativeExecutionEngine(ExecutionEngine):
     def repartition(
         self, df: DataFrame, partition_spec: PartitionSpec
     ) -> DataFrame:  # pragma: no cover
-        self.log.warning(f"{self} doesn't respect repartition")
+        self.log.warning("%s doesn't respect repartition", self)
         return df
 
     def map(
@@ -107,7 +137,9 @@ class NativeExecutionEngine(ExecutionEngine):
     ) -> DataFrame:
         if partition_spec.num_partitions != "0":
             self.log.warning(
-                f"{self} doesn't respect num_partitions {partition_spec.num_partitions}"
+                "%s doesn't respect num_partitions %s",
+                self,
+                partition_spec.num_partitions,
             )
         cursor = partition_spec.get_cursor(df.schema, 0)
         if on_init is not None:
@@ -116,9 +148,15 @@ class NativeExecutionEngine(ExecutionEngine):
             df = to_local_df(df)
             cursor.set(df.peek_array(), 0, 0)
             output_df = map_func(cursor, df)
+            if (
+                isinstance(output_df, PandasDataFrame)
+                and output_df.schema != output_schema
+            ):
+                output_df = PandasDataFrame(output_df.native, output_schema)
             assert_or_throw(
                 output_df.schema == output_schema,
-                f"map output {output_df.schema} mismatches given {output_schema}",
+                lambda: f"map output {output_df.schema} "
+                f"mismatches given {output_schema}",
             )
             output_df._metadata = ParamDict(metadata, deep=True)
             output_df._metadata.set_readonly()
@@ -176,7 +214,8 @@ class NativeExecutionEngine(ExecutionEngine):
         metadata: Any = None,
     ) -> DataFrame:
         assert_or_throw(
-            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
+            df1.schema == df2.schema,
+            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
         )
         d = self.pl_utils.union(df1.as_pandas(), df2.as_pandas(), unique=distinct)
         return PandasDataFrame(d.reset_index(drop=True), df1.schema, metadata)
@@ -192,7 +231,8 @@ class NativeExecutionEngine(ExecutionEngine):
             distinct, NotImplementedError("EXCEPT ALL for NativeExecutionEngine")
         )
         assert_or_throw(
-            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
+            df1.schema == df2.schema,
+            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
         )
         d = self.pl_utils.except_df(df1.as_pandas(), df2.as_pandas(), unique=distinct)
         return PandasDataFrame(d.reset_index(drop=True), df1.schema, metadata)
@@ -208,7 +248,8 @@ class NativeExecutionEngine(ExecutionEngine):
             distinct, NotImplementedError("INTERSECT ALL for NativeExecutionEngine")
         )
         assert_or_throw(
-            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
+            df1.schema == df2.schema,
+            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
         )
         d = self.pl_utils.intersect(df1.as_pandas(), df2.as_pandas(), unique=distinct)
         return PandasDataFrame(d.reset_index(drop=True), df1.schema, metadata)
@@ -224,16 +265,97 @@ class NativeExecutionEngine(ExecutionEngine):
     def dropna(
         self,
         df: DataFrame,
-        metadata: Any = None,
         how: str = "any",
-        thresh: int = None,
+        thresh: Optional[int] = None,
         subset: List[str] = None,
+        metadata: Any = None,
     ) -> DataFrame:
-        d = df.as_pandas().dropna(
-            axis=0, how=how, thresh=thresh, subset=subset, inplace=False
-        )
+        kwargs: Dict[str, Any] = dict(axis=0, subset=subset, inplace=False)
+        if thresh is None:
+            kwargs["how"] = how
+        else:
+            kwargs["thresh"] = thresh
+        d = df.as_pandas().dropna(**kwargs)
         return PandasDataFrame(d.reset_index(drop=True), df.schema, metadata)
 
+    def fillna(
+        self,
+        df: DataFrame,
+        value: Any,
+        subset: List[str] = None,
+        metadata: Any = None,
+    ) -> DataFrame:
+        assert_or_throw(
+            (not isinstance(value, list)) and (value is not None),
+            ValueError("fillna value can not None or a list"),
+        )
+        if isinstance(value, dict):
+            assert_or_throw(
+                (None not in value.values()) and (any(value.values())),
+                ValueError(
+                    "fillna dict can not contain None and needs at least one value"
+                ),
+            )
+            mapping = value
+        else:
+            # If subset is none, apply to all columns
+            subset = subset or df.schema.names
+            mapping = {col: value for col in subset}
+        d = df.as_pandas().fillna(mapping, inplace=False)
+        return PandasDataFrame(d.reset_index(drop=True), df.schema, metadata)
+
+    def sample(
+        self,
+        df: DataFrame,
+        n: Optional[int] = None,
+        frac: Optional[float] = None,
+        replace: bool = False,
+        seed: Optional[int] = None,
+        metadata: Any = None,
+    ) -> DataFrame:
+        assert_or_throw(
+            (n is None and frac is not None) or (n is not None and frac is None),
+            ValueError("one and only one of n and frac should be set"),
+        )
+        d = df.as_pandas().sample(n=n, frac=frac, replace=replace, random_state=seed)
+        return PandasDataFrame(d.reset_index(drop=True), df.schema, metadata)
+
+    def take(
+        self,
+        df: DataFrame,
+        n: int,
+        presort: str,
+        na_position: str = "last",
+        partition_spec: PartitionSpec = EMPTY_PARTITION_SPEC,
+        metadata: Any = None,
+    ) -> DataFrame:
+        assert_or_throw(
+            isinstance(n, int),
+            ValueError("n needs to be an integer"),
+        )
+        d = df.as_pandas()
+
+        # Use presort over partition_spec.presort if possible
+        if presort:
+            presort = parse_presort_exp(presort)
+        _presort: IndexedOrderedDict = presort or partition_spec.presort
+
+        if len(_presort.keys()) > 0:
+            d = d.sort_values(
+                list(_presort.keys()),
+                ascending=list(_presort.values()),
+                na_position=na_position,
+            )
+
+        if len(partition_spec.partition_by) == 0:
+            d = d.head(n)
+        else:
+            d = d.groupby(by=partition_spec.partition_by, dropna=False).head(n)
+
+        return PandasDataFrame(
+            d.reset_index(drop=True), df.schema, metadata, pandas_df_wrapper=True
+        )
+
     def load_df(
         self,
         path: Union[str, List[str]],
@@ -257,9 +379,27 @@ class NativeExecutionEngine(ExecutionEngine):
         force_single: bool = False,
         **kwargs: Any,
     ) -> None:
-        if not partition_spec.empty:
-            self.log.warning(  # pragma: no cover
-                f"partition_spec is not respected in {self}.save_df"
-            )
+        if not force_single and not partition_spec.empty:
+            kwargs["partition_cols"] = partition_spec.partition_by
+        self.fs.makedirs(os.path.dirname(path), recreate=True)
         df = self.to_df(df)
         save_df(df, path, format_hint=format_hint, mode=mode, fs=self.fs, **kwargs)
+
+
+class _NativeExecutionEngineParam(ExecutionEngineParam):
+    def __init__(
+        self,
+        param: Optional[inspect.Parameter],
+    ):
+        super().__init__(
+            param, annotation="NativeExecutionEngine", engine_type=NativeExecutionEngine
+        )
+
+
+register_annotation_converter(
+    0.8,
+    SimpleAnnotationConverter(
+        NativeExecutionEngine,
+        lambda param: _NativeExecutionEngineParam(param),
+    ),
+)
