@@ -1,21 +1,30 @@
 # flake8: noqa
-from fugue.extensions._builtins.creators import CreateData, Load
+from fugue.extensions._builtins.creators import Load, LoadYielded
 from fugue.extensions._builtins.outputters import (
     AssertEqual,
     AssertNotEqual,
+    RunOutputTransformer,
     Save,
     Show,
 )
 from fugue.extensions._builtins.processors import (
+    Aggregate,
+    AlterColumns,
+    Assign,
     Distinct,
     DropColumns,
     Dropna,
+    Fillna,
+    Filter,
     Rename,
     RunJoin,
     RunSetOperation,
     RunSQLSelect,
     RunTransformer,
+    Sample,
     SaveAndUse,
+    Select,
     SelectColumns,
+    Take,
     Zip,
 )
