@@ -1,13 +1,14 @@
-import os
 from typing import Any
 
-from fugue.dataframe import DataFrame
+import fs as pfs
+from fugue.collections.partition import PartitionSpec
+from fugue.collections.yielded import YieldedFile
 from fugue.constants import FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH
-from fugue.exceptions import FugueWorkflowError
+from fugue.dataframe import DataFrame
+from fugue.exceptions import FugueWorkflowCompileError, FugueWorkflowRuntimeError
 from fugue.execution.execution_engine import ExecutionEngine
 from triad.utils.assertion import assert_or_throw
 from triad.utils.hash import to_uuid
-from fugue.collections.partition import PartitionSpec
 
 
 class Checkpoint(object):
@@ -15,30 +16,36 @@ class Checkpoint(object):
         self,
         to_file: bool = False,
         deterministic: bool = False,
+        permanent: bool = False,
         lazy: bool = False,
         **kwargs: Any,
     ):
+        if deterministic:
+            assert_or_throw(permanent, "deterministic checkpoint must be permanent")
         self.to_file = to_file
         self.deterministic = deterministic
+        self.permanent = permanent
         self.lazy = lazy
         self.kwargs = dict(kwargs)
 
-    def run(self, file_id: str, df: DataFrame, path: "CheckpointPath") -> DataFrame:
+    def run(self, df: DataFrame, path: "CheckpointPath") -> DataFrame:
         return df
 
-    def __uuid__(self) -> str:
-        return to_uuid(
-            self.to_file,
-            self.deterministic,
-            self.lazy,
-            self.kwargs,
-        )
+    @property
+    def yielded_file(self) -> YieldedFile:
+        raise FugueWorkflowCompileError(f"yield file is not allowed for {self}")
+
+    @property
+    def is_null(self) -> bool:
+        return True
 
 
 class FileCheckpoint(Checkpoint):
     def __init__(
         self,
+        file_id: str,
         deterministic: bool,
+        permanent: bool,
         lazy: bool = False,
         partition: Any = None,
         single: bool = False,
@@ -48,6 +55,7 @@ class FileCheckpoint(Checkpoint):
         super().__init__(
             to_file=True,
             deterministic=deterministic,
+            permanent=permanent,
             lazy=lazy,
             fmt="",
             partition=PartitionSpec(partition),
@@ -55,9 +63,12 @@ class FileCheckpoint(Checkpoint):
             namespace=namespace,
             save_kwargs=dict(save_kwargs),
         )
+        self._yield_func: Any = None
+        self._file_id = to_uuid(file_id, namespace)
+        self._yielded = YieldedFile(self._file_id)
 
-    def run(self, file_id: str, df: DataFrame, path: "CheckpointPath") -> DataFrame:
-        fpath = path.get_temp_file(file_id, self.deterministic)
+    def run(self, df: DataFrame, path: "CheckpointPath") -> DataFrame:
+        fpath = path.get_temp_file(self._file_id, self.permanent)
         if not self.deterministic or not path.temp_file_exists(fpath):
             path.execution_engine.save_df(
                 df=df,
@@ -68,7 +79,23 @@ class FileCheckpoint(Checkpoint):
                 force_single=self.kwargs["single"],
                 **self.kwargs["save_kwargs"],
             )
-        return path.execution_engine.load_df(path=fpath, format_hint=self.kwargs["fmt"])
+        result = path.execution_engine.load_df(
+            path=fpath, format_hint=self.kwargs["fmt"]
+        )
+        self._yielded.set_value(fpath)
+        return result
+
+    @property
+    def yielded_file(self) -> YieldedFile:
+        assert_or_throw(
+            self.permanent,
+            lambda: FugueWorkflowCompileError(f"yield is not allowed for {self}"),
+        )
+        return self._yielded
+
+    @property
+    def is_null(self) -> bool:
+        return False
 
 
 class WeakCheckpoint(Checkpoint):
@@ -84,9 +111,13 @@ class WeakCheckpoint(Checkpoint):
             **kwargs,
         )
 
-    def run(self, file_id: str, df: DataFrame, path: "CheckpointPath") -> DataFrame:
+    def run(self, df: DataFrame, path: "CheckpointPath") -> DataFrame:
         return path.execution_engine.persist(df, lazy=self.lazy, **self.kwargs)
 
+    @property
+    def is_null(self) -> bool:
+        return False
+
 
 class CheckpointPath(object):
     def __init__(self, engine: ExecutionEngine):
@@ -104,7 +135,7 @@ class CheckpointPath(object):
         if self._path == "":
             self._temp_path = ""
             return ""
-        self._temp_path = os.path.join(self._path, execution_id)
+        self._temp_path = pfs.path.combine(self._path, execution_id)
         self._fs.makedirs(self._temp_path, recreate=True)
         return self._temp_path
 
@@ -113,15 +144,17 @@ class CheckpointPath(object):
             try:
                 self._fs.removetree(self._temp_path)
             except Exception as e:  # pragma: no cover
-                self._log.warn("Unable to remove " + self._temp_path, e)
+                self._log.info("Unable to remove " + self._temp_path, e)
 
-    def get_temp_file(self, file_id: str, deterministic: bool) -> str:
-        path = self._path if deterministic else self._temp_path
+    def get_temp_file(self, file_id: str, permanent: bool) -> str:
+        path = self._path if permanent else self._temp_path
         assert_or_throw(
             path != "",
-            FugueWorkflowError(f"{FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH} is not set"),
+            FugueWorkflowRuntimeError(
+                f"{FUGUE_CONF_WORKFLOW_CHECKPOINT_PATH} is not set"
+            ),
         )
-        return os.path.join(path, file_id + ".parquet")
+        return pfs.path.combine(path, file_id + ".parquet")
 
     def temp_file_exists(self, path: str) -> bool:
         try:
