@@ -1,14 +1,20 @@
 import logging
-from typing import Any, Callable, List, Optional, Union
+import os
+from typing import Any, Callable, Dict, List, Optional, Union
 
-import dask.dataframe as pd
-from fugue._utils.io import load_df, save_df
+import dask.dataframe as dd
+from distributed import Client
 from fugue.collections.partition import (
     EMPTY_PARTITION_SPEC,
     PartitionCursor,
     PartitionSpec,
+    parse_presort_exp,
+)
+from fugue.constants import (
+    FUGUE_CONF_SQL_IGNORE_CASE,
+    KEYWORD_CORECOUNT,
+    KEYWORD_ROWCOUNT,
 )
-from fugue.constants import KEYWORD_CORECOUNT, KEYWORD_ROWCOUNT
 from fugue.dataframe import DataFrame, DataFrames, LocalDataFrame, PandasDataFrame
 from fugue.dataframe.utils import get_join_schemas
 from fugue.execution.execution_engine import (
@@ -16,19 +22,22 @@ from fugue.execution.execution_engine import (
     ExecutionEngine,
     SQLEngine,
 )
+from fugue.execution.native_execution_engine import NativeExecutionEngine
 from qpd_dask import run_sql_on_dask
 from triad.collections import Schema
-from triad.collections.dict import ParamDict
+from triad.collections.dict import IndexedOrderedDict, ParamDict
 from triad.collections.fs import FileSystem
 from triad.utils.assertion import assert_or_throw
 from triad.utils.hash import to_uuid
 from triad.utils.threading import RunOnce
 
 from fugue_dask._constants import (
+    CPU_COUNT,
     FUGUE_DASK_CONF_DATAFRAME_DEFAULT_PARTITIONS,
     FUGUE_DASK_DEFAULT_CONF,
 )
-from fugue_dask._utils import DaskUtils
+from fugue_dask._io import load_df, save_df
+from fugue_dask._utils import DASK_UTILS, DaskUtils
 from fugue_dask.dataframe import DaskDataFrame
 
 
@@ -39,6 +48,10 @@ class QPDDaskEngine(SQLEngine):
     """
 
     def __init__(self, execution_engine: ExecutionEngine):
+        assert_or_throw(
+            isinstance(execution_engine, DaskExecutionEngine),
+            lambda: f"{self} must be used with DaskExecutionEngine",
+        )
         super().__init__(execution_engine)
 
     def select(self, dfs: DataFrames, statement: str) -> DataFrame:
@@ -46,7 +59,13 @@ class QPDDaskEngine(SQLEngine):
             k: self.execution_engine.to_df(v).native  # type: ignore
             for k, v in dfs.items()
         }
-        df = run_sql_on_dask(statement, dask_dfs)
+        df = run_sql_on_dask(
+            statement,
+            dask_dfs,
+            ignore_case=self.execution_engine.compile_conf.get(
+                FUGUE_CONF_SQL_IGNORE_CASE, False
+            ),
+        )
         return DaskDataFrame(df)
 
 
@@ -55,26 +74,36 @@ class DaskExecutionEngine(ExecutionEngine):
 
     Please read |ExecutionEngineTutorial| to understand this important Fugue concept
 
+    :param dask_client: Dask distributed client, defaults to None. If None, then it
+      will try to get the current active global client. If there is no active client,
+      it will create and use a global `Client(processes=True)`
     :param conf: |ParamsLikeObject| defaults to None, read |FugueConfig| to
       learn Fugue specific options
 
-    :Notice:
+    .. note::
 
-    You should setup Dask single machine or distributed environment in the
-    :doc:`common <dask:setup>` way. Before initializing :class:`~.DaskExecutionEngine`
+        You should setup Dask single machine or distributed environment in the
+        :doc:`common <dask:setup>` way.
+        Before initializing :class:`~.DaskExecutionEngine`
     """
 
-    def __init__(self, conf: Any = None):
+    def __init__(self, dask_client: Optional[Client] = None, conf: Any = None):
         p = ParamDict(FUGUE_DASK_DEFAULT_CONF)
         p.update(ParamDict(conf))
         super().__init__(p)
         self._fs = FileSystem()
         self._log = logging.getLogger()
-        self._default_sql_engine = QPDDaskEngine(self)
+        self._client = DASK_UTILS.get_or_create_client(dask_client)
+        self._native = NativeExecutionEngine(conf=conf)
 
     def __repr__(self) -> str:
         return "DaskExecutionEngine"
 
+    @property
+    def dask_client(self) -> Client:
+        """The Dask Client associated with this engine"""
+        return self._client
+
     @property
     def log(self) -> logging.Logger:
         return self._log
@@ -85,17 +114,13 @@ class DaskExecutionEngine(ExecutionEngine):
 
     @property
     def default_sql_engine(self) -> SQLEngine:
-        return self._default_sql_engine
+        return QPDDaskEngine(self)
 
     @property
     def pl_utils(self) -> DaskUtils:
         """Pandas-like dataframe utils"""
         return DaskUtils()
 
-    def stop(self) -> None:  # pragma: no cover
-        """It does nothing"""
-        return
-
     def to_df(self, df: Any, schema: Any = None, metadata: Any = None) -> DaskDataFrame:
         """Convert a data structure to :class:`~fugue_dask.dataframe.DaskDataFrame`
 
@@ -106,15 +131,15 @@ class DaskExecutionEngine(ExecutionEngine):
         :param metadata: |ParamsLikeObject|, defaults to None
         :return: engine compatible dataframe
 
-        :Notice:
+        .. note::
 
-        * if the input is already :class:`~fugue_dask.dataframe.DaskDataFrame`,
-          it should return itself
-        * For list or iterable of arrays, ``schema`` must be specified
-        * When ``schema`` is not None, a potential type cast may happen to ensure
-          the dataframe's schema.
-        * all other methods in the engine can take arbitrary dataframes and
-          call this method to convert before doing anything
+            * if the input is already :class:`~fugue_dask.dataframe.DaskDataFrame`,
+              it should return itself
+            * For list or iterable of arrays, ``schema`` must be specified
+            * When ``schema`` is not None, a potential type cast may happen to ensure
+              the dataframe's schema.
+            * all other methods in the engine can take arbitrary dataframes and
+              call this method to convert before doing anything
         """
         default_partitions = self.conf.get_or_throw(
             FUGUE_DASK_CONF_DATAFRAME_DEFAULT_PARTITIONS, int
@@ -149,15 +174,17 @@ class DaskExecutionEngine(ExecutionEngine):
         p = partition_spec.get_num_partitions(
             **{
                 KEYWORD_ROWCOUNT: lambda: df.persist().count(),  # type: ignore
-                KEYWORD_CORECOUNT: lambda: 2,  # TODO: remove this hard code
+                KEYWORD_CORECOUNT: lambda: CPU_COUNT,
             }
         )
         if p > 0:
+            if partition_spec.algo == "even":
+                pdf = df.as_pandas()
+                ddf = dd.from_pandas(pdf, npartitions=p, sort=False)
+            else:
+                ddf = df.native.repartition(npartitions=p)
             return DaskDataFrame(
-                df.native.repartition(npartitions=p),
-                schema=df.schema,
-                metadata=df.metadata,
-                type_safe=False,
+                ddf, schema=df.schema, metadata=df.metadata, type_safe=False
             )
         return df
 
@@ -183,7 +210,7 @@ class DaskExecutionEngine(ExecutionEngine):
             )
         )
 
-        def _map(pdf: Any) -> pd.DataFrame:
+        def _map(pdf: Any) -> dd.DataFrame:
             if pdf.shape[0] == 0:
                 return PandasDataFrame([], output_schema).as_pandas()
             if len(presort_keys) > 0:
@@ -199,16 +226,14 @@ class DaskExecutionEngine(ExecutionEngine):
             return output_df.as_pandas()
 
         df = self.to_df(df)
+        meta = self.pl_utils.safe_to_pandas_dtype(output_schema.pa_schema)
         if len(partition_spec.partition_by) == 0:
             pdf = self.repartition(df, partition_spec)
-            result = pdf.native.map_partitions(_map, meta=output_schema.pandas_dtype)
+            result = pdf.native.map_partitions(_map, meta=meta)
         else:
             df = self.repartition(df, PartitionSpec(num=partition_spec.num_partitions))
             result = self.pl_utils.safe_groupby_apply(
-                df.native,
-                partition_spec.partition_by,
-                _map,
-                meta=output_schema.pandas_dtype,
+                df.native, partition_spec.partition_by, _map, meta=meta
             )
         return DaskDataFrame(result, output_schema, metadata)
 
@@ -248,7 +273,8 @@ class DaskExecutionEngine(ExecutionEngine):
         metadata: Any = None,
     ) -> DataFrame:
         assert_or_throw(
-            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
+            df1.schema == df2.schema,
+            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
         )
         d = self.pl_utils.union(
             self.to_df(df1).native, self.to_df(df2).native, unique=distinct
@@ -266,7 +292,8 @@ class DaskExecutionEngine(ExecutionEngine):
             distinct, NotImplementedError("EXCEPT ALL for DaskExecutionEngine")
         )
         assert_or_throw(
-            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
+            df1.schema == df2.schema,
+            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
         )
         d = self.pl_utils.except_df(
             self.to_df(df1).native, self.to_df(df2).native, unique=distinct
@@ -284,7 +311,8 @@ class DaskExecutionEngine(ExecutionEngine):
             distinct, NotImplementedError("INTERSECT ALL for DaskExecutionEngine")
         )
         assert_or_throw(
-            df1.schema == df2.schema, ValueError(f"{df1.schema} != {df2.schema}")
+            df1.schema == df2.schema,
+            lambda: ValueError(f"{df1.schema} != {df2.schema}"),
         )
         d = self.pl_utils.intersect(
             self.to_df(df1).native, self.to_df(df2).native, unique=distinct
@@ -302,12 +330,120 @@ class DaskExecutionEngine(ExecutionEngine):
     def dropna(
         self,
         df: DataFrame,
-        metadata: Any = None,
         how: str = "any",
         thresh: int = None,
         subset: List[str] = None,
+        metadata: Any = None,
     ) -> DataFrame:
-        d = self.to_df(df).native.dropna(how=how, thresh=thresh, subset=subset)
+        kw: Dict[str, Any] = dict(how=how)
+        if thresh is not None:
+            kw["thresh"] = thresh
+        if subset is not None:
+            kw["subset"] = subset
+        if how == "any" and thresh is not None:
+            del kw["how"]  # to deal with a dask logic flaw
+        d = self.to_df(df).native.dropna(**kw)
+        return DaskDataFrame(d, df.schema, metadata)
+
+    def fillna(
+        self,
+        df: DataFrame,
+        value: Any,
+        subset: List[str] = None,
+        metadata: Any = None,
+    ) -> DataFrame:
+        assert_or_throw(
+            (not isinstance(value, list)) and (value is not None),
+            ValueError("fillna value can not be a list or None"),
+        )
+        if isinstance(value, dict):
+            assert_or_throw(
+                (None not in value.values()) and (any(value.values())),
+                ValueError(
+                    "fillna dict can not contain None and needs at least one value"
+                ),
+            )
+            mapping = value
+        else:
+            # If subset is none, apply to all columns
+            subset = subset or df.schema.names
+            mapping = {col: value for col in subset}
+        d = self.to_df(df).native.fillna(mapping)
+        return DaskDataFrame(d, df.schema, metadata)
+
+    def sample(
+        self,
+        df: DataFrame,
+        n: Optional[int] = None,
+        frac: Optional[float] = None,
+        replace: bool = False,
+        seed: Optional[int] = None,
+        metadata: Any = None,
+    ) -> DataFrame:
+        assert_or_throw(
+            (n is None and frac is not None) or (n is not None and frac is None),
+            ValueError("one and only one of n and frac should be set"),
+        )
+        # TODO: dask does not support sample by number of rows
+        d = self.to_df(df).native.sample(
+            n=n, frac=frac, replace=replace, random_state=seed
+        )
+        return DaskDataFrame(d, df.schema, metadata)
+
+    def take(
+        self,
+        df: DataFrame,
+        n: int,
+        presort: str,
+        na_position: str = "last",
+        partition_spec: PartitionSpec = EMPTY_PARTITION_SPEC,
+        metadata: Any = None,
+    ) -> DataFrame:
+        assert_or_throw(
+            isinstance(n, int),
+            ValueError("n needs to be an integer"),
+        )
+        d = self.to_df(df).native
+        meta = [(d[x].name, d[x].dtype) for x in d.columns]
+
+        if presort:
+            presort = parse_presort_exp(presort)
+        # Use presort over partition_spec.presort if possible
+        _presort: IndexedOrderedDict = presort or partition_spec.presort
+
+        def _partition_take(partition, n, presort):
+            if len(presort.keys()) > 0:
+                partition = partition.sort_values(
+                    list(presort.keys()),
+                    ascending=list(presort.values()),
+                    na_position=na_position,
+                )
+            return partition.head(n)
+
+        if len(partition_spec.partition_by) == 0:
+            if len(_presort.keys()) == 0:
+                d = d.head(n)
+            else:
+                # Use the default partition
+                d = (
+                    d.map_partitions(_partition_take, n, _presort, meta=meta)
+                    .reset_index(drop=True)
+                    .compute()
+                )
+                # compute() brings this to Pandas so we can use pandas
+                d = d.sort_values(
+                    list(_presort.keys()),
+                    ascending=list(_presort.values()),
+                    na_position=na_position,
+                ).head(n)
+
+        else:
+            d = (
+                d.groupby(partition_spec.partition_by, dropna=False)
+                .apply(_partition_take, n=n, presort=_presort, meta=meta)
+                .reset_index(drop=True)
+            )
+
         return DaskDataFrame(d, df.schema, metadata)
 
     def load_df(
@@ -333,9 +469,19 @@ class DaskExecutionEngine(ExecutionEngine):
         force_single: bool = False,
         **kwargs: Any,
     ) -> None:
-        if not partition_spec.empty:
-            self.log.warning(  # pragma: no cover
-                f"partition_spec is not respected in {self}.save_df"
+        if force_single:
+            self._native.save_df(
+                df,
+                path=path,
+                format_hint=format_hint,
+                mode=mode,
+                partition_spec=partition_spec,
+                force_single=force_single,
+                **kwargs,
             )
-        df = self.to_df(df).as_local()
-        save_df(df, path, format_hint=format_hint, mode=mode, fs=self.fs, **kwargs)
+        else:
+            if not partition_spec.empty:
+                kwargs["partition_on"] = partition_spec.partition_by
+            self.fs.makedirs(os.path.dirname(path), recreate=True)
+            df = self.to_df(df)
+            save_df(df, path, format_hint=format_hint, mode=mode, fs=self.fs, **kwargs)
