@@ -1,10 +1,37 @@
 import json
 
-from fugue.collections.partition import PartitionSpec
+from fugue.collections.partition import parse_presort_exp, PartitionSpec
 from fugue.constants import KEYWORD_CORECOUNT, KEYWORD_ROWCOUNT
 from pytest import raises
 from triad.collections.schema import Schema
 from triad.utils.hash import to_uuid
+from triad.collections.dict import IndexedOrderedDict
+
+def test_parse_presort_exp():
+
+    assert parse_presort_exp(None) == IndexedOrderedDict()
+    assert parse_presort_exp(IndexedOrderedDict([('c', True)])) == IndexedOrderedDict([('c', True)])
+    assert parse_presort_exp("c") == IndexedOrderedDict([('c', True)])
+    assert parse_presort_exp("         c") == IndexedOrderedDict([('c', True)])
+    assert parse_presort_exp("c           desc")  == IndexedOrderedDict([('c', False)])
+    assert parse_presort_exp("b desc, c asc")  == IndexedOrderedDict([('b', False), ('c', True)])
+    assert parse_presort_exp("DESC DESC, ASC ASC") == IndexedOrderedDict([('DESC', False), ('ASC', True)])
+    assert parse_presort_exp([("b", False),("c", True)]) == IndexedOrderedDict([('b', False), ('c', True)])
+    assert parse_presort_exp("B DESC, C ASC")  == IndexedOrderedDict([('B', False), ('C', True)])
+    assert parse_presort_exp("b desc, c asc") == IndexedOrderedDict([('b', False), ('c', True)])
+    
+
+    with raises(SyntaxError):
+        parse_presort_exp("b dsc, c asc") # mispelling of desc
+
+    with raises(SyntaxError):
+        parse_presort_exp("c true") # string format needs desc/asc
+
+    with raises(SyntaxError):
+        parse_presort_exp("c true, c true") # cannot contain duplicates
+
+    with raises(SyntaxError):
+        parse_presort_exp([("b", "desc"),("c", "asc")]) # instead of desc and asc, needs to be bool
 
 
 def test_partition_spec():
@@ -41,17 +68,32 @@ def test_partition_spec():
     assert "even" == p.algo
     assert not p.empty
 
-    p = PartitionSpec(partition_by=["a", "b", "c"], presort="d,e desc", algo="EvEN",
-                      num_partitions="ROWCOUNT*3", row_limit=4, size_limit="5k")
+    p = PartitionSpec(
+        partition_by=["a", "b", "c"],
+        presort="d,e desc",
+        algo="EvEN",
+        num_partitions="ROWCOUNT*3",
+        row_limit=4,
+        size_limit="5k",
+    )
     p2 = PartitionSpec(p)
     assert p2.jsondict == p.jsondict
     assert "d ASC,e DESC" == p2.presort_expr
     assert not p.empty
     assert not p2.empty
+    print(p)
+    print(f"{p}")
+
+    assert PartitionSpec("per_row") == PartitionSpec(num="ROWCOUNT", algo="even")
+    assert PartitionSpec(by="abc") == PartitionSpec(by=["abc"])
 
     # partition by overlaps with presort
-    raises(SyntaxError, lambda: PartitionSpec(partition_by=[
-           "a", "b", "c"], presort="a asc,e desc", algo="EvEN"))
+    raises(
+        SyntaxError,
+        lambda: PartitionSpec(
+            partition_by=["a", "b", "c"], presort="a asc,e desc", algo="EvEN"
+        ),
+    )
 
     # partition by has dups
     raises(SyntaxError, lambda: PartitionSpec(partition_by=["a", "b", "b"]))
@@ -59,6 +101,9 @@ def test_partition_spec():
     # partition by has dups
     raises(SyntaxError, lambda: PartitionSpec(partition_by=["a", "b", "b"]))
 
+    # partition by is invalid
+    raises(SyntaxError, lambda: PartitionSpec(partition_by=123))
+
     # bad input
     raises(TypeError, lambda: PartitionSpec(1))
 
@@ -68,12 +113,13 @@ def test_partition_spec():
     raises(SyntaxError, lambda: PartitionSpec(presort="a b asc,a desc"))
     raises(SyntaxError, lambda: PartitionSpec(presort=["a asc", ("b", True)]))
     raises(SyntaxError, lambda: PartitionSpec(presort=[("a", "asc"), "b"]))
-    raises(SyntaxError, lambda: PartitionSpec(presort=[("a", ), ("b")]))
+    raises(SyntaxError, lambda: PartitionSpec(presort=[("a",), ("b")]))
     raises(SyntaxError, lambda: PartitionSpec(presort=["a", ["b", True]]))
 
     p = PartitionSpec(dict(partition_by=["a"], presort="d asc,e desc"))
     assert dict(a=True, d=True, e=False) == p.get_sorts(
-        Schema("a:int,b:int,d:int,e:int"))
+        Schema("a:int,b:int,d:int,e:int")
+    )
     p = PartitionSpec(dict(partition_by=["e", "a"], presort="d asc"))
     assert p.get_key_schema(Schema("a:int,b:int,d:int,e:int")) == "e:int,a:int"
 
@@ -81,9 +127,9 @@ def test_partition_spec():
     a = PartitionSpec(by=["a", "b"])
     b = PartitionSpec(a, by=["a"], num=2)
     assert ["a", "b"] == a.partition_by
-    assert '0' == a.num_partitions
+    assert "0" == a.num_partitions
     assert ["a"] == b.partition_by
-    assert '2' == b.num_partitions
+    assert "2" == b.num_partitions
 
     a = PartitionSpec(by=["a"], presort="b DESC, c")
     b = PartitionSpec(by=["a"], presort="c,b DESC")
@@ -101,7 +147,7 @@ def test_partition_spec():
     assert b.presort == c.presort
     assert a.presort == d.presort
 
-    a = PartitionSpec(by=["a"], presort=["b","c"])
+    a = PartitionSpec(by=["a"], presort=["b", "c"])
     b = PartitionSpec(by=["a"], presort=[("b"), ("c")])
     c = PartitionSpec(by=["a"], presort=[("b", True), ("c", True)])
     d = PartitionSpec(by=["a"], presort="b asc, c")
@@ -109,6 +155,17 @@ def test_partition_spec():
     assert a.presort == c.presort
     assert a.presort == d.presort
 
+    # test eq
+    a = PartitionSpec(by=["a"], presort=["b", "c"])
+    b = PartitionSpec(presort="b, c asc", by=["a"])
+    c = PartitionSpec(num=10, by=["a"], presort=["b", "c"])
+    d = PartitionSpec(num=10, by=["a"], presort=["c", "b"])
+    assert a == a
+    assert a == dict(presort=["b", "c"], by=["a"])
+    assert a == b
+    assert a != c
+    assert c != d
+
 
 def test_partition_cursor():
     p = PartitionSpec(dict(partition_by=["b", "a"]))
@@ -141,7 +198,8 @@ def test_get_num_partitions():
 
     p = PartitionSpec(dict(partition_by=["b", "a"], num="min(ROWCOUNT,CORECOUNT)"))
     assert 90 == p.get_num_partitions(
-        **{KEYWORD_ROWCOUNT: lambda: 100, KEYWORD_CORECOUNT: lambda: 90})
+        **{KEYWORD_ROWCOUNT: lambda: 100, KEYWORD_CORECOUNT: lambda: 90}
+    )
 
 
 def test_determinism():
