@@ -9,11 +9,14 @@ import pandas
 from fugue.dataframe.array_dataframe import ArrayDataFrame
 from fugue.dataframe.pandas_dataframe import PandasDataFrame
 from fugue.dataframe.utils import _df_eq as df_eq
-from fugue.exceptions import FugueDataFrameInitError
 from fugue_dask.dataframe import DaskDataFrame
 from fugue_test.dataframe_suite import DataFrameTests
 from pytest import raises
 from triad.collections.schema import Schema
+from fugue.dataframe.utils import (
+    get_dataframe_column_names,
+    rename_dataframe_column_names,
+)
 
 
 class DaskDataFrameTests(DataFrameTests.Tests):
@@ -30,7 +33,7 @@ def test_init():
     assert df.schema == "a:str,b:int"
 
     pdf = pandas.DataFrame([["a", 1], ["b", 2]])
-    raises(FugueDataFrameInitError, lambda: DaskDataFrame(pdf))
+    raises(Exception, lambda: DaskDataFrame(pdf))
     df = DaskDataFrame(pdf, "a:str,b:str")
     assert [["a", "1"], ["b", "2"]] == df.as_pandas().values.tolist()
     df = DaskDataFrame(pdf, "a:str,b:int")
@@ -60,7 +63,7 @@ def test_init():
     df = DaskDataFrame([], "x:str,y:double")
     assert [] == df.as_pandas().values.tolist()
 
-    raises(FugueDataFrameInitError, lambda: DaskDataFrame(123))
+    raises(Exception, lambda: DaskDataFrame(123))
 
 
 def test_simple_methods():
@@ -129,7 +132,7 @@ def test_as_array():
 
     df = DaskDataFrame([[pandas.NaT, 1.1]], "a:datetime,b:int")
     df.native["a"] = pd.to_datetime(df.native["a"])
-    assert isinstance(df.as_array()[0][0], datetime)
+    assert df.as_array()[0][0] is None
     assert isinstance(df.as_array()[0][1], int)
 
     df = DaskDataFrame([[1.0, 1.1]], "a:double,b:int")
@@ -159,11 +162,11 @@ def test_nan_none():
     df = ArrayDataFrame([["a", 1.1], [None, None]], "b:str,c:double")
     arr = DaskDataFrame(df.as_pandas(), df.schema).as_array()[1]
     assert arr[0] is None
-    assert math.isnan(arr[1])
+    assert arr[1] is None
 
     arr = DaskDataFrame(df.as_array(), df.schema).as_array()[1]
     assert arr[0] is None
-    assert math.isnan(arr[1])
+    assert arr[1] is None
 
     arr = DaskDataFrame(df.as_pandas()["b"], "b:str").as_array()[1]
     assert arr[0] is None
@@ -196,3 +199,22 @@ def _test_as_array_perf():
         res = df.as_array(type_safe=True)
         ts += (datetime.now() - t).total_seconds()
     print(nts, ts)
+
+
+def test_get_dataframe_column_names():
+    df = pd.from_pandas(pandas.DataFrame([[0, 1, 2]]), npartitions=1)
+    assert get_dataframe_column_names(df) == [0, 1, 2]
+
+
+def test_rename_dataframe_column_names():
+    pdf = pd.from_pandas(
+        pandas.DataFrame([[0, 1, 2]], columns=["a", "b", "c"]), npartitions=1
+    )
+    df = rename_dataframe_column_names(pdf, {})
+    assert isinstance(df, pd.DataFrame)
+    assert get_dataframe_column_names(df) == ["a", "b", "c"]
+
+    pdf = pd.from_pandas(pandas.DataFrame([[0, 1, 2]]), npartitions=1)
+    df = rename_dataframe_column_names(pdf, {0: "_0", 1: "_1", 2: "_2"})
+    assert isinstance(df, pd.DataFrame)
+    assert get_dataframe_column_names(df) == ["_0", "_1", "_2"]
