@@ -4,15 +4,19 @@ from fugue.collections.partition import PartitionSpec
 from fugue.dataframe import DataFrame, DataFrames, LocalDataFrame
 from fugue.workflow.workflow import FugueWorkflow
 from fugue_sql.exceptions import FugueSQLSyntaxError
-from fugue_sql._parse import FugueSQL
+from fugue_sql_antlr import FugueSQLParser
 from fugue_sql._visitors import _Extensions, _VisitorBase
 from pytest import raises
 from triad.collections.schema import Schema
 
+_PARSE_MODE = "auto"
+
 
 def test_json():
     def assert_eq(expr, expected):
-        sql = FugueSQL(expr, "fugueJsonValue", ignore_case=True)
+        sql = FugueSQLParser(
+            expr, "fugueJsonValue", ignore_case=True, parse_mode=_PARSE_MODE
+        )
         v = _VisitorBase(sql)
         obj = v.visit(sql.tree)
         if expected is None:
@@ -50,7 +54,9 @@ def test_json():
 
 def test_schema():
     def assert_eq(expr, expected=None):
-        sql = FugueSQL(expr, "fugueSchema", ignore_case=True)
+        sql = FugueSQLParser(
+            expr, "fugueSchema", ignore_case=True, parse_mode=_PARSE_MODE
+        )
         v = _VisitorBase(sql)
         obj = v.visit(sql.tree)
         if expected is None:
@@ -62,11 +68,14 @@ def test_schema():
     assert_eq("a:int,B:str,c:[int]")
     assert_eq("a:int,b:str,C:{A:int,b:str}")
     assert_eq("a:int,\nb:str,C:[{A:int,b:str}]")
+    assert_eq("a:<int,str>,b:<int,[str]>,c:[<int,{b:str}>]")
 
 
 def test_wild_schema():
     def assert_eq(expr, expected=None):
-        sql = FugueSQL(expr, "fugueWildSchema", ignore_case=True)
+        sql = FugueSQLParser(
+            expr, "fugueWildSchema", ignore_case=True, parse_mode=_PARSE_MODE
+        )
         v = _VisitorBase(sql)
         obj = v.visit(sql.tree)
         if expected is None:
@@ -79,13 +88,20 @@ def test_wild_schema():
     assert_eq("*  ", "*")
     assert_eq("*,a:int32", "*,a:int")
     assert_eq("a:int32, *, \nb:string", "a:int,*,b:str")
+    # transformations
+    assert_eq(
+        "a:int32, *, B:string +x:str, k:int\n-Y, z -t~ w , x",
+        "a:int,*,B:str+x:str,k:int-Y,z-t~w,x",
+    )
     with raises(FugueSQLSyntaxError):
         assert_eq("*,a:int32, *, \nb:string")
 
 
 def test_pre_partition():
     def assert_eq(expr, expected):
-        sql = FugueSQL(expr, "fuguePrepartition", ignore_case=True)
+        sql = FugueSQLParser(
+            expr, "fuguePrepartition", ignore_case=True, parse_mode=_PARSE_MODE
+        )
         v = _VisitorBase(sql)
         obj = json.dumps(v.visit(sql.tree).jsondict)
         assert json.dumps(expected.jsondict) == obj
@@ -116,7 +132,9 @@ def test_pre_partition():
 
 def test_params():
     def assert_eq(expr, expected):
-        sql = FugueSQL(expr, "fugueParams", ignore_case=True)
+        sql = FugueSQLParser(
+            expr, "fugueParams", ignore_case=True, parse_mode=_PARSE_MODE
+        )
         v = _VisitorBase(sql)
         obj = v.visit(sql.tree)
         assert expected == obj
@@ -130,10 +148,15 @@ def test_params():
 
 def test_single_output_common_expr():
     def assert_eq(expr, using, params, schema):
-        sql = FugueSQL(expr, "fugueSingleOutputExtensionCommon", ignore_case=True)
+        sql = FugueSQLParser(
+            expr,
+            "fugueSingleOutputExtensionCommon",
+            ignore_case=True,
+            parse_mode=_PARSE_MODE,
+        )
         v = _VisitorBase(sql)
         obj = v.visit(sql.tree)
-        assert using == obj["using"]
+        assert using == obj["fugueUsing"]
         if params is None:
             assert "params" not in obj
         else:
@@ -157,11 +180,12 @@ def test_single_output_common_expr():
 
 def test_assignment():
     def assert_eq(expr, varname, sign):
-        sql = FugueSQL(expr, "fugueAssignment", ignore_case=True, simple_assign=True)
+        sql = FugueSQLParser(
+            expr, "fugueAssignment", ignore_case=True, parse_mode=_PARSE_MODE
+        )
         v = _VisitorBase(sql)
         obj = v.visit(sql.tree)
         assert (varname, sign) == obj
 
-    assert_eq("aA:=", "aA", ":=")
     assert_eq("aA=", "aA", "=")
     # assert_eq("aA??", "aA", "??")
